/*

This include file provides the structures to efficiently store a
block-diagonal-bordered sparse matrix to be used in a Gauss-Seidel
algorithm.  This implicit sparse data structure is static.  This data
structure will be used for either sequential or parallel applications.

In this data structure, diagonal matrix blocks and the associated
portions of the lower border and right border are identified
separately.  There are six portions to this implicit sparse matrix
data storage methodology:

   1) Block identifier

   2) Matrix diagonal

   3) Block row storage

   4) Lower border

   5) Lower border row storage

   These data structures are designed for run-time allocation of memory,
   so there is no wasted memory due to static allocation at compile time.

   These data structures do not explicitly include the number of non-zero
   elements in each row or column portion - rather this information is
   available by comparing pointers to the previous row/column and the
   pointer to the present row/column.  In fact, the data stored is the pointer
   to the last data in a row/column.  These values can be directly used
   in the looping structure to minimize indexing overhead in both the
   sequential and parallel algorithms.  Blocks, etc., will be numbered
   starting at 1, so the previous location will be the "zeroth" vector
   slot for the structure.  All vectors of structures will require an
   additional slot to faciliate simple use of this data structure.  All
   values of the zeroth slot will be initialized to zero.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <malloc.h>
#include <math.h>
#include <cm/cmmd.h>
#include "interface.h"
#include "message.h"
#include "sets.h"

#ifndef GS_H
#define GS_H

#define YES        1
#define NO         0
#define EPSILON    0.0000001

typedef struct ProcObj
{
   int             BlockPtr;    /* proc[] */
   int             DiagPtr;
   int             BlockRowPtr;
   int             LowerBorderPtr;
   int             LowerBorderRowPtr;
   Set            *LBSet;
}               Proc;

typedef struct BlockObj
{
   int             BlockID;     /* blocks[] */
   int             BlockSz;
   int             ProcNum;
   int             DiagPtr;
}               Block;

typedef struct DiagObj
{
   double          Value;       /* diag[] */
   int             PtrAix;
   Set            *BlockRowSet;
}               Diag;

typedef struct BlockRowObj
{
   double          Value;       /* blockrow[] */
   int             i;
   int             j;
}               BlockRow;

typedef struct LowerBorderObj
{
   int             i;           /* lborder[] */
   int             ulbproc;
   int             ulbloc;
   int             PtrLBix;
   Set            *LBorderRowSet;
}               LowerBorder;

typedef struct LowerBorderRowObj
{
   double          Value;       /* lborderrow[] */
   int             i;
   int             j;
}               LBorderRow;

/* data structures for last diagonal block */

typedef struct LastBlockColorObj
{
   int             Color;       /* lbcolor[icolor][iproc] */
   int             ProcNum;
   int             Starti;
   int             DiagPtr
}               LastBlockColor;

typedef struct LastDiagObj
{
   double          Value;       /* ldiag[] */
   int             PtrAix;
}               LastDiag;

typedef struct LastBlockRowObj
{
   double          Value;       /* lblockrow[] */
   int             i;
   int             j;
}               LastBlockRow;

/* data structures for the updates of the last block using data in
   the borders --- different data structure for each proc */

typedef struct UpdateLastBlockObj
{
   double          Value;       /* updatelb[iproc][] */
   int             i;
}               UpdateLastBlock;

/* data structures for communications when solving for nodes in
   the last block */

typedef struct CommLBObj
{
   int             starti;      /* commlb[icolor][iproc] */
   int             bufsize;
}               CommLB;

/*---------------------------------------------------------*/
typedef struct DBLObj
{
   unsigned int    i1;
   unsigned int    i2;
}               DBL;
typedef struct RESObj
{
   int             proc;
   double          res;
}               RES;

#endif

void            InputData(char *, char *);
int             min(int, int);

void            GSSolveTiming_1();
void            GSSolve_1();
void            GS_1();
void            GSConverge_1();

void            GSSolveTiming();
void            GSSolve();
void            GSDiagonalBlocks();
void            GSUpdateLastBlock();
void            GSUpdateLastBlockAsync();
void            GSUpdateLastBlockCMAML();
void            StoreHandlerULBCMAML();
void            GSLastBlock();
void            SendHandlerCMMD();
void            ReceiveHandlerCMMD();
void            GSLastBlockCMAML();
void            StoreHandlerLBCMAML();
void            GSConverge();
void            GSConvergeCMAML();
void            StoreHandlerConCMAML();
void            Node_Exit();


